#include "pico/stdlib.h"
#include <stdio.h>
#include "hardware/pwm.h"
#include "hardware/clocks.h"
#include "hardware/adc.h"
#include "../xiao_rp2040.h"

// #include "sound.h"

#define NR_SAMPLES_SOUND 144000

#define PIN_SPEAKER D10
#define PIN_PLAY D3
#define PIN_RECORD D5

// the pin we are using for analog
#define INPUT_GUITAR 0
// the pin for the first port
#define PIN_ANALOG A0 + INPUT_GUITAR

#define SAMPLE_RATE 48000
#define OVERSAMPLING_FACTOR 20

#define NOTE_HZ 400

#define BITS_ADC 12
#define BITS_SPEAKER 7

//uint8_t sound[NR_SAMPLES_SOUND];

uint16_t compute_wrap() {
  return (uint16_t) (clock_get_hz(clk_sys) / (SAMPLE_RATE * OVERSAMPLING_FACTOR));
}


// a wrap will occur 960000 times per second
// to get a tone of 400 Hz, it should turn on and off every 48000 * 20 / 400
void on_pwm_wrap_dim() {

  static int nr_samples = 0;
  static bool high = true;
  static int level = 0;
  static int high_level = 0;
  static bool going_up = true;
  static int period = 0;

  // Clear the interrupt flag that brought us here
  pwm_clear_irq(pwm_gpio_to_slice_num(PIN_SPEAKER));

  if (period == 3200) {
    period = 0;
    if (going_up) {
      high_level++;
      if (high_level > 128) {
	high_level = 128;
	going_up = false;
      }
    }
    else {
      high_level--;
      if (high_level < 16) {
	high_level = 16;
	going_up = true;
      }
    }
  }
  else {
    period++;
  }

  if (nr_samples == (SAMPLE_RATE * OVERSAMPLING_FACTOR / 2 / NOTE_HZ)) {
    // set the number of samples to zero
    nr_samples = 0;

    // set the level
    pwm_set_gpio_level(PIN_SPEAKER, level);
    if (high) {
      level = high_level;
      high = false;
    }
    else {
      level = 0;
      high = true;
    }
  }
  else {
    nr_samples++;
  }
}

// a wrap will occur 960000 times per second
void on_pwm_wrap_sound() {

  static int recording = 1;
  static int nr_wraps = 0;
  static int sample = 0;

  // Clear the interrupt flag that brought us here
  pwm_clear_irq(pwm_gpio_to_slice_num(PIN_SPEAKER));

  if (nr_wraps == 0) { // this happens once every 48000 times
    if (recording) {
      uint adc_raw = adc_read();
      sound[sample] = adc_raw >> 4;
    }
    else {
      // set the level
      pwm_set_gpio_level(PIN_SPEAKER, sound[sample]/2);
    }
    sample++;
  }

  if (nr_wraps == OVERSAMPLING_FACTOR - 1) {
    // set the number of wraps to zero
    nr_wraps = 0;
  }
  else {
    nr_wraps++;
  }

  if (sample == NR_SAMPLES_SOUND - 1) {
    sample = 0;
    recording = !recording;
    if (recording) {
      pwm_set_gpio_level(PIN_SPEAKER, 0);
      gpio_put(PIN_RECORD, 1);
      gpio_put(PIN_PLAY, 0);
    }
    else {
      gpio_put(PIN_RECORD, 0);
      gpio_put(PIN_PLAY, 1);
    }
  }
}


int main() {
  stdio_init_all();

  adc_init();
  adc_gpio_init(PIN_ANALOG);
  adc_select_input(INPUT_GUITAR);

  gpio_init(PIN_PLAY);
  gpio_set_dir(PIN_PLAY, GPIO_OUT);

  gpio_init(PIN_RECORD);
  gpio_set_dir(PIN_RECORD, GPIO_OUT);
  
  // Tell the pin that the PWM is in charge of its value.
  gpio_set_function(PIN_SPEAKER, GPIO_FUNC_PWM);

  // Figure out which slice we just connected to the LED pin
  uint slice_num = pwm_gpio_to_slice_num(PIN_SPEAKER);

  pwm_clear_irq(slice_num);
  pwm_set_irq_enabled(slice_num, true);
  irq_set_exclusive_handler(PWM_IRQ_WRAP, on_pwm_wrap_dim);
  irq_set_enabled(PWM_IRQ_WRAP, true);
  
  // Get some sensible defaults for the slice configuration. By default, the
  // counter is allowed to wrap over its maximum range (0 to 2**16-1)
  pwm_config config = pwm_get_default_config();

  // set the wrap value a little bit higher than 128, so we can use this number
  // and we will never reach a 100% duty cycle
  uint16_t wrap = compute_wrap();
  
  pwm_config_set_wrap(&config, wrap);
  pwm_init(slice_num, &config, true);

  sleep_ms(3000);

  printf("system clock: %d\n", clock_get_hz(clk_sys));
  printf("sample rate: %d\n", SAMPLE_RATE);
  printf("oversampling factor: %d\n", OVERSAMPLING_FACTOR);
  printf("wrap: %d\n", wrap);

  
  while (1) {
    tight_loop_contents();
  }
}
